"""
Definition of optimization problems information Pydantic classes.

Created Nov. 14, 2024

@author Joerg Gablonsky
"""

import re
from typing import List, Optional, Dict, Tuple, Union, Literal, Any
import typing
import numpy as np
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    model_validator,
    field_serializer,
)
from numpydantic import NDArray, Shape
from standard_evaluator import unique_names

MAXINT = 2**63


# Generating names of array variables
# We want to generate variable / response names that reflect that we have an array.
def generate_names(name: str, shape: tuple) -> typing.List[str]:
    """Generate names for variables / responses that reflect that they are arrays

    The names generated use NumPy array syntax. For example, a variable input that
    is a 1-d array of length 2 (shape (2,)) will generate this list:

    ['input[0]', 'input[1]']

    Parameters
    ----------
    name : str
        Name of the variable or response
    shape : tuple
        Shape of the variable or response. This is a NumPy shape

    Returns
    -------
    list[str]
        List of all the variables / responses represented by the array

    Raises
    ------
    TypeError
        If the name is not a string, or the shape is not a tuple
    """
    if not isinstance(name, str):
        raise TypeError("Name is not a string")
    if not isinstance(shape, tuple):
        raise TypeError("Shape is not a tuple")

    new_names = []
    for idx in np.ndindex(shape):
        new_names.append(f"{name}[" + ",".join([str(value) for value in idx]) + "]")
    return new_names


class FloatVariable(BaseModel):
    name: str = Field(description="Name of the variable.")
    default: Optional[float] = Field(
        default=None, description="Default value for this variable"
    )
    bounds: Optional[Tuple[float, float]] = Field(
        default=(-np.inf, np.inf), description="Lower and upper bounds of the variable"
    )
    shift: Optional[float] = Field(
        default=0.0,
        description="Shift value to be used for this variable.",
    )
    scale: Optional[float] = Field(
        default=1.0,
        description="Scale value to be used for this variable." + "Cannot be zero.",
    )
    units: Optional[str] = None
    description: Optional[str] = Field(
        default="",
        description="A description of the variable.",
    )
    options: Dict = Field(default_factory=dict, description="Options for the variable.")
    class_type: Literal["float"] = Field(default="float", description="Class marker")

    @field_validator("bounds")
    def validate_bounds(cls, var):
        if var is None:
            return (-np.inf, np.inf)
        if var[0] > var[1]:
            raise ValueError(
                f"Lower bounds are larger than upper bounds: {var[0]} > {var[1]}"
            )
        return var

    @field_validator("scale")
    def validate_scale(cls, scale):
        if scale == 0:
            raise ValueError(f"Scale cannot be 0.")
        return scale

    @field_validator("name")
    def check_name(cls, name: str) -> str:
        # We want to make sure the following checks are only done when
        # the name is not an empty string
        if len(name) > 0:
            if re.search(r"\s", name):
                raise ValueError(f"Name cannot contain white spaces.")
        return name


class IntVariable(FloatVariable):
    default: Optional[int] = Field(
        default=None, description="Default value for this variable"
    )
    bounds: Optional[Tuple[int, int]] = Field(
        default=[-MAXINT, MAXINT], description="Lower and upper bounds of the variable"
    )
    shift: Optional[int] = Field(
        default=0,
        description="Shift value to be used for this variable.",
    )
    scale: Optional[int] = Field(
        default=1,
        description="Scale value to be used for this variable." + "Cannot be zero.",
    )
    class_type: Literal["int"] = Field(default="int", description="Class marker")

    @field_validator("bounds")
    def validate_bounds(cls, var):
        if var is None:
            return (-MAXINT, MAXINT)
        if var[0] > var[1]:
            raise ValueError(
                f"Lower bounds are larger than upper bounds: {var[0]} > {var[1]}"
            )
        return var


class CategoricalVariable(FloatVariable):
    default: Optional[Any] = Field(
        default=None, description="Default value for this variable"
    )
    bounds: List[Any] = Field(
        description="Values this variable or response can take on. Must be defined"
    )
    shift: Literal[None] = Field(
        default=None,
        description="Shift value to be used for this variable. Does not make sense for categorical variables",
    )
    scale: Literal[None] = Field(
        default=None,
        description="Scale value to be used for this variable. Does not make sense for categorical variables",
    )
    units: Optional[str] = None
    class_type: Literal["cat"] = Field(default="cat", description="Class marker")

    @field_validator("bounds")
    def validate_bounds(cls, var):
        """Method to overwrite the inherited method. For categorical variables bounds can be"""
        if len(var) == 0:
            raise ValueError(
                f"Need to define at least one valid value for this variable"
            )
        return var

    @model_validator(mode="after")
    def check_default(self):
        """Method to chek that the default value, if it is defined, is valid"""
        bounds = self.bounds
        default = self.default
        if default is not None:
            if default not in bounds:
                raise ValueError(f"Default not in bounds: {default}")
        return self


class ArrayVariable(FloatVariable):
    """Class defining array variables. The underlying data type are NumPy float64 arrays

    Raises:
        ValueError: Lower bounds greater than upper bounds for at least one element
        ValueError: Neither shape or default are set
        ValueError: Scale for at least one element is set to 0.
        ValueError: Inconsistent shapes for elements in the class
    """

    shape: Optional[Tuple[int, ...]] = Field(
        default=None, description="Shape of the arrays"
    )
    default: Optional[NDArray[Shape["*,..."], np.float64]] = Field(
        default=None,
        description="NumPy array with the default values to be used for the array variable",
    )
    bounds: Optional[
        Tuple[
            Union[float, int, NDArray[Shape["*,..."], np.float64]],
            Union[float, int, NDArray[Shape["*,..."], np.float64]],
        ]
    ] = Field(default=(None, None), description="Lower and upper bounds of the arrays")
    shift: Optional[Union[float, int, NDArray[Shape["*,..."], np.float64]]] = Field(
        default=None,
        description="NumPy array with the shift values to be used for the array variable.",
    )
    scale: Optional[Union[float, int, NDArray[Shape["*,..."], np.float64]]] = Field(
        default=None,
        description="NumPy array with the scale values to be used for the array variable. "
        + "Cannot be zero.",
    )
    class_type: Literal["floatarray"] = "floatarray"

    @field_validator("bounds")
    def validate_bounds(cls, var):
        if var is None:
            return (None, None)
        if np.any(var[0] > var[1]):
            raise ValueError(
                f"Lower bounds are larger than upper bounds: {var[0]} > {var[1]}"
            )
        return var

    @field_validator("scale")
    def validate_scale(cls, scale):
        # Because we only want to do this validation once we have a shape defined it
        # happens in the check_shape method
        return scale

    def _expand(self, value, scaler=1.0):
        if value is not None:
            if isinstance(value, (float, int)):
                value = np.ones(self.shape) * value
        else:
            value = np.ones(self.shape) * scaler
        return value

    @model_validator(mode="after")
    def check_shape(self):
        shape = self.shape
        default = self.default
        shift = self.shift
        scale = self.scale

        if shape is None:
            # If no shape is defined  default must be set. If default is
            # set it's shape will define the shape. If neither is set we throw an error
            if default is not None:
                shape = default.shape
            else:
                raise ValueError(
                    f"Shape for this variable is not set, and neither are default or value"
                )
            self.shape = shape

        # If shift is set we want to check if only a single value is set. If that is the
        # case we expand it to the full shape.
        shift = self._expand(shift, 0.0)
        self.shift = shift

        # If scale is set we want to check if only a single value is set. If that is the
        # case we expand it to the full shape.
        scale = self._expand(scale)

        # We have to make sure none of the values of scale are 0.
        if np.any(scale == 0.0):
            raise ValueError(f"Scale for one element in the matrix is set to 0.0")

        self.scale = scale

        # Check that the bounds are set correctly
        (low_bound, up_bound) = self.bounds
        low_bound = self._expand(low_bound, -np.inf)
        up_bound = self._expand(up_bound, np.inf)
        self.bounds = (low_bound, up_bound)

        fields = [default, shift, scale, low_bound, up_bound]
        field_names = ["default", "shift", "scale", "lower bound", "upper bound"]
        # Check if default is not None. If default is set we check that the shape is
        # set correctly.
        for local_field, field_name in zip(fields, field_names):
            if local_field is not None:
                if local_field.shape != shape:
                    raise ValueError(
                        f"Shape for this variable is {shape}, but {field_name} has shape {local_field.shape}"
                    )

        return self

    # If the bounds are set to (-inf, inf) we don't need to store them.
    @field_serializer("bounds")
    def serialize_bounds(
        self,
        bounds: Tuple[
            NDArray[Shape["*,..."], np.float64],
            NDArray[Shape["*,..."], np.float64],
        ],
    ):
        if np.all(np.isinf(bounds[0])) & np.all(np.isinf(bounds[0])):
            return None
        else:
            lower = bounds[0]
            upper = bounds[1]
            if np.all(upper == upper.ravel()[0]):
                upper = upper.ravel()[0]
            if np.all(lower == lower.ravel()[0]):
                lower = lower.ravel()[0]
            bounds = (lower, upper)
            return bounds

    # If shift is set to all zeros we don't need to save them
    @field_serializer("shift")
    def serialize_shift(self, shift: NDArray[Shape["*,..."], np.float64]):
        if np.all(shift == 0.0):
            return None
        else:
            return shift

    # If scale is set to all ones we don't need to save them
    @field_serializer("scale")
    def serialize_scale(self, scale: NDArray[Shape["*,..."], np.float64]):
        if np.all(scale == 1.0):
            return None
        else:
            return scale


# Define the Union of the different variable types. Note that we use that for responses as well
Variable = Union[FloatVariable, IntVariable, ArrayVariable, CategoricalVariable]


class OptProblem(BaseModel):
    """
    Represents information about an optimization problem.

    Args:
        name: The name of the problem. Defaults to 'opt_problem'
        variables: A list of input variables.
        responses: A list of output variables.
        objectives: Names of the objective(s) for the optimization problem. Must be either variables or responses defined in the problem..
        constraints: Names of the constraints of the optimization problem. Must be responses defined in the problem. To define bounds on variables use the variable bounds.
        description: A description of the optimization problem. To define mathematical symbols use markdown syntax.
        cite: Listing of relevant citations that should be referenced when publishing work that uses this class.
        options: Additional options for the problem.
    """

    name: str = Field(
        default="opt_problem",
        description='The name of the problem. Defaults to "opt_problem"',
    )
    class_type: Literal["OptProblem"] = "OptProblem"
    variables: List[Variable] = Field(description="Input variables")
    responses: List[Variable] = Field(description="Output variables")
    objectives: List[str] = Field(
        default_factory=list,
        description="Names of the objective(s) for the optimization problem. Must be either variables or responses defined in the problem.",
    )
    constraints: List[str] = Field(
        default_factory=list,
        description="Names of the constraints of the optimization problem. Must be responses defined in the problem. To define bounds on variables use the variable bounds.",
    )
    description: Optional[str] = Field(
        default=None,
        description="A description of the optimization problem. To define mathematical symbols use markdown syntax.",
    )
    cite: Optional[str] = Field(
        default=None,
        description="Listing of relevant citations that should be referenced when publishing work that uses this class.",
    )
    options: Dict = Field(
        default_factory=dict, description="Additional options for the problem."
    )

    @field_validator("variables", "responses")
    def validate_outputs(cls, var):
        unique_names(var)
        return var

    def unroll_names(self, elements: List[Variable]) -> List[str]:
        all_names = []
        for local_element in elements:
            name = local_element.name
            if isinstance(local_element, ArrayVariable):
                array_names = generate_names(name, local_element.shape)
                all_names += array_names
            else:
                all_names.append(name)
        return all_names

    @model_validator(mode="after")
    def check_problem(self):

        # This method needs to be updated to allow using an element from a FloatArray
        variable_names = self.unroll_names(self.variables)
        response_names = self.unroll_names(self.responses)
        elements = set(variable_names + response_names)
        if self.objectives is not None:
            # Check if all the objectives are either a variable or response
            for name in self.objectives:
                if name not in elements:
                    raise ValueError(
                        f"{name} is defined as an objective, but not defined as a variable or response."
                    )

        if self.constraints is not None:
            # Check if all the objectives are either a variable or response
            for name in self.constraints:
                if name not in response_names:
                    if name in variable_names:
                        raise ValueError(
                            f"{name} is defined as a constraint, but defined as a variable. Please define bounds on the variable itself. Constraints should only be responses."
                        )
                    raise ValueError(
                        f"{name} is defined as a constraint, but not defined as a response."
                    )

        return self



""" Implementing the Standard Evaluator for components    """

import inspect
import typing

from standard_evaluator import StandardBase
from aviary.variable_info.variable_meta_data import _MetaData


class StandardEval(StandardBase):
    """A class to be used in conjunction with OpenMDAO components to have a common way of dealing with options.

    This class implements the ability of the system to keep track of options,
    and also pre-defines that this component depends on the 'class_options'.
    """

    def initialize(self):
        """Initialize the instance by defining that this will use `class_options`, 
        and 'metadata', and set the 'metadata' options to `_MetaData` by default."""
        super().initialize()
        # Define an option to set the meta data. By default it will be set to `_MetaData` from
        # Aviary
        self.options.declare(
            "metadata",
            types=dict,
            desc="A dictionary containing metadata and default values to use for variables or responses",
        )
        self.options["metadata"] = _MetaData

    def _lookup_info(
        self,
        name: str,
        val: typing.Optional[typing.Any] = None,
        units: typing.Optional[str] = None,
        desc: typing.Optional[str] = None,
    ) -> typing.Tuple[typing.Any, str, str]:
        """Helper method to look up information based on the meta data.

        Arguments:
            name {str} -- Name of the element to look up information for

        Keyword Arguments:
            val {typing.Optional[typing.Any]} -- Value to use. If None value will be looked up (default: {None})
            units {typing.Optional[str]} --Units to use. If None units will be looked up (default: {None})
            desc {typing.Optional[str]} -- Description to use. If None description will be looked up (default: {None})

        Returns:
            typing.Tuple[typing.Any, str, str] -- Updated values, units, and description. 
        """
        meta = self.options["metadata"][name]
        if units is None:
            units = meta["units"]
        if desc is None:
            desc = meta["desc"]
        if val is None:
            val = meta["default_value"]
        return (val, units, desc)



""" Implementing the Standard Evaluator for groups

Groups are different in that they have components within them. When we find out what all the
options for the group we have to take into account the options for all the components in the
group.
    """

import inspect
import typing

from standard_evaluator import StandardBase, OptionsDictionaryUnit, opt_dict_unit_merge

class StandardGroup(StandardBase):
    """A class to be used in conjunction with OpenMDAO groups to have a common way of dealing with options.

    This class implements the ability of the system to keep track of options,
    and also pre-defines that this component depends on the 'class_options'.
    """

    def initialize(self):
        """Initialize the instance by defining that this will use `class_options`, 
        and 'metadata', and set the 'metadata' options to `_MetaData` by default."""
        super().initialize()

    def full_options(self) -> OptionsDictionaryUnit:
        """An instance method to return all options for this class.
        
        Often this will be just a call to the `required_options` class method. But for example
        in group the full set of options that are valid are depending on the components in the
        instance, so this has to be calculated after an instance is created, and the `setup()`
        method has been called.

        Returns:
            OptionsDictionaryUnit -- All valid options for this class.
        """
        
        options_list = []
        for subsys in self._subsystems_myproc:
            if isinstance(subsys, StandardBase):
                options_list.append(subsys.full_options())
        
        return opt_dict_unit_merge(options_list)

    def current_options(self) -> OptionsDictionaryUnit:
        """Return all the current options. 
        This is a combination between the original options (defaults) and values that
        were overwritten by the end user.

        Returns:
            OptionsDictionaryUnit -- The combination of all options
        """
        options_list = []
        for subsys in self._subsystems_myproc:
            if isinstance(subsys, StandardBase):
                options_list.append(subsys.current_options())
        
        return opt_dict_unit_merge(options_list)

    def get_interface(self) -> dict:
        """Generate a dictionary with all the information known about this component

        Returns:
            dict: Dictionary with all information known about the component
        """
        info = super().get_interface()

        info["components"] = []
        # Since this is a group we need to collect the information for all subsystems
        for subsys in self._subsystems_myproc:
            if isinstance(subsys, StandardBase):
                info["components"].append(subsys.get_interface())
        info["general"]["group"] = True

        return info



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import importlib
import inspect
import os
import re
import sys
from datetime import date

sys.path.insert(0, os.path.abspath("../../"))
sys.path.append(os.path.abspath('./_ext'))


# -- Project information -----------------------------------------------------

project = "Standard Evaluator"
copyright = f"None"
author = "Joerg Gablonsky, Jeff Musiak, Ranald M Engelbeck, Eduardo Ocampo, Sean Wakayama"

# Version information

from pkg_resources import parse_version
from standard_evaluator import version

version_text = version.__version__
v = parse_version(version_text)

# The short X.Y version
version = v.base_version

# The full version, including alpha / beta / rc tags
release = version_text


# -- General configuration ---------------------------------------------------


def setup(app):
    pass


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # 'sphinx.ext.duration',
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "autodocsumm",
    # Adding the ability to use Mermaid diagrams
    "sphinxcontrib.mermaid",
    # Custom directive to print options and known solution
    # Adds :print-options: and :known-solution: options to autoclass
    'format-class',
    "myst_nb",
    # This is needed to prevent a weird lexer unknown error when converting
    # jupyter notebooks to html
    'IPython.sphinxext.ipython_console_highlighting',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Place documentation generated by automodule and autoclass in the order they
# are in the source file.
autodoc_member_order = "bysource"

# Make it so that classes will not look like funtions
# ie.
#   class ClassName
# instead of
#   class ClassName(arg1, arg2, arg3)
autodoc_class_signature = "separated"

# Parameter types will only be in the description. Thus making things cleaner
autodoc_typehints = "description"

autodoc_default_options = {
    "autosummary": True,
    "autosummary-sections": "Attributes ;; Methods",
    "members": True,
    "special-members": True,
    "show-inheritance": True,
    "exclude-members": "__weakref__",
}

# Jupyter notebook cell executation time limit. Use None for no time limit
nb_execution_timeout = 600

# Hides class names from table of contents and sidebar
# When True it will look like:
#   - Python Evaluator
#       - PyEvaluator
#   - Executable Evaluator
#       - ExecutableEvaluator
#
# When False it will look like:
#   - Python Evaluator
#   - Executable Evaluator
toc_object_entries = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Options for RTD theme
# https://sphinx-rtd-theme.readthedocs.io/en/stable/configuring.html
html_theme_options = {
    'prev_next_buttons_location': 'None'
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Location of custom css files. Must be relative to html_static_path
html_css_files = [
    "css/class_doc.css",
    'css/general.css',
]

# Location of custom js files. Must be relative to html_static_path
html_js_files = [
    'js/require.min.js',
    'js/custom.js',
]

# Latex Support

katex_options = r"""{
    delimiters: [
        {left: '$$', right: '$$', display: true},
        {left: '$', right: '$', display: false},
        {left: '\\(', right: '\\)', display: false},
        {left: '\\[', right: '\\]', display: true}
    ],
}"""

# Mermaid options
mermaid_d3_zoom = True

# Myst-NB options

jupyter_execute_notebooks = "cache"
execution_in_temp = False
jupyter_cache = os.path.join(os.path.dirname(__file__), ".cache")
if not os.path.exists(jupyter_cache):
    os.mkdir(jupyter_cache)

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "replacements",
    "smartquotes",
    "strikethrough",
    "tasklist",
]

# tweaks for how we write LaTeX

# allow spaces around $ delimiters
myst_dmath_allow_space = True

# allow displaymath in a markdown 'inline' context.
myst_dmath_double_inline = True

# hide module names
add_module_names = False


"""
Test harness for `test_system_description_to_json.py`.
"""
import logging
import numpy as np
import unittest
import os
from openmdao.core.constants import _UNDEFINED
import json

# for pipeline
from standard_evaluator import system_description_to_json as sd
from standard_evaluator import circuit_example as ce


# Module attributes.  ----------------------------------------------------

DEBUG = False
"""Whether to log debugging information and/or execute debugging code."""

RUN_ALL_TESTS = True
"""Whether to run all tests or manually selected tests."""


# Unit test classes.  ----------------------------------------------------

# class TestSystemDescriptionToJson(unittest.TestCase):
class test_system_description_to_json(unittest.TestCase):

    # Infrastructure methods.  -------------------------------------------

    def setUp(self):
        # Define test problem and model
        self.cir = ce.Circuit()
        self.cir.create_model()

        # run n2 to genereate the basic dict
        self.test_dict_basic = sd.model_to_dict(self.cir.p, values=True, title="JDM")

    # Unit test methods.  ------------------------------------------------

    @unittest.skipIf(not RUN_ALL_TESTS, "Temporarily disabled.")
    def test_get_additional_options(self):
        """
        Test `get_additional_options()`.
        """
        ###
        # Configuration.
        test_dict = {"val": 3, "recordable": True}
        add_dict = {"n1": {"additional_option_1": "little", "additional_option_2": "big"},
                    "n2": {"additional_option_1": "little", "additional_option_2": "big"},
                    "D1": {"additional_option_1": "little", "additional_option_2": "big"},
                    "R1": {"additional_option_1": "little", "additional_option_2": "big"}}
        out_dict = {'Optional': {'D1': {'additional_option_1': 'little',
                                        'additional_option_2': 'big'},
                                 'R1': {'additional_option_1': 'little',
                                        'additional_option_2': 'big'},
                                 'n1': {'additional_option_1': 'little',
                                        'additional_option_2': 'big'},
                                 'n2': {'additional_option_1': 'little',
                                        'additional_option_2': 'big'}},
                    'recordable': True,
                    'val': 3}
        add_option_file_ = "additional_options.json"
        with open(add_option_file_, 'w') as adF:
            json.dump(add_dict, adF)

        # Execution.
        test_add_opt = sd.get_additional_options(test_dict, add_option_file_)

        # Validation.
        assert test_add_opt == out_dict

    # TODO: Can't figure out how to get this to work
    @unittest.skipIf(not RUN_ALL_TESTS, "Temporarily disabled.")
    def test__serialize_single_option_big(self):
        """
        Test `serialize_single_option()`.
        """
        ###
        # Configuration.
        test_dict = {"val": "From Wikipedia, the free encyclopedia This article is about Homer's epic poem. For other uses, see Odyssey (disambiguation). 'Homer's Odyssey' redirects here. For The Simpsons episode, see Homer's Odyssey (The Simpsons). Odyssey by Homer 15th-century manuscript of Book I written by scribe John Rhosos (British Museum) Written	c. 8th century BC Language	Homeric Greek Genre(s)	Epic poetry Publisher	1488 Published in English	1614 Lines	12,109 Metre	Dactylic hexameter Full text The Odyssey at Wikisource  The Odyssey (/ˈɒdɪsi/;[1] Ancient Greek: Ὀδύσσεια, romanized: Odýsseia)[2][3] is one of two major ancient Greek epic poems attributed to Homer. It is one of the oldest extant works of literature still widely read by modern audiences. As with the Iliad, the poem is divided into 24 books. It follows the Greek hero Odysseus, king of Ithaca, and his journey home after the Trojan War. After the war, which lasted ten years, his journey from Troy to Ithaca, via Africa and southern Europe, lasted for ten additional years during which time he encountered many perils and all of his crewmates were killed. In his absence, Odysseus was assumed dead, and his wife Penelope and son Telemachus had to contend with a group of unruly suitors who were competing for Penelope's hand in marriage.  The Odyssey was originally composed in Homeric Greek in around the 8th or 7th century BC and, by the mid-6th century BC, had become part of the Greek literary canon. In antiquity, Homer's authorship of the poem was not questioned, but contemporary scholarship predominantly assumes that the Iliad and the Odyssey were composed independently and that the stories formed as part of a long oral tradition. Given widespread illiteracy, the poem was performed by an aoidos or rhapsode and was more likely to be heard than read.  Crucial themes in the poem include the ideas of nostos (νόστος; 'return'), wandering, xenia (ξενία; 'guest-friendship'), testing, and omens. Scholars still reflect on the narrative significance of certain groups in the poem, such as women and slaves, who have a more prominent role in the epic than in many other works of ancient literature. This focus is especially remarkable when contrasted with the Iliad, which centres the exploits of soldiers and kings during the Trojan War.  The Odyssey is regarded as one of the most significant works of the Western canon. The first English translation of the Odyssey was in the 16th century. Adaptations and re-imaginings continue to be produced across a wide variety of media. In 2018, when BBC Culture polled experts around the world to find literature's most enduring narrative, the Odyssey topped the list.[4] Synopsis Exposition (books 1–4) A mosaic depicting Odysseus, from the villa of La Olmeda, Pedrosa de la Vega, Spain, late 4th–5th centuries AD  The Odyssey begins after the end of the ten-year Trojan War (the subject of the Iliad), from which Odysseus (also known by the Latin variant Ulysses), king of Ithaca, has still not returned because he angered Poseidon, the god of the sea. Odysseus' son, Telemachus, is about 20 years old and is sharing his absent father's house on the island of Ithaca with his mother Penelope and the suitors of Penelope, a crowd of 108 boisterous young men who each aim to persuade Penelope for her hand in marriage, all the while reveling in the king's palace and eating up his wealth.  Odysseus' protectress, the goddess Athena, asks Zeus, king of the gods, to finally allow Odysseus to return home when Poseidon is absent from Mount Olympus. Disguised as a chieftain named Mentes, Athena visits Telemachus to urge him to search for news of his father. He offers her hospitality, and they observe the suitors dining rowdily while Phemius, the bard, performs a narrative poem for them.  That night, Athena, disguised as Telemachus, finds a ship and crew for the true prince. The next morning, Telemachus calls an assembly of citizens of Ithaca to discuss what should be done with the insolent suitors, who then scoff at Telemachus. Accompanied by Athena (now disguised as Mentor), the son of Odysseus departs for the Greek mainland to the household of Nestor, most venerable of the Greek warriors at Troy, who resided in Pylos after the war.  From there, Telemachus rides to Sparta, accompanied by Nestor's son. There he finds Menelaus and Helen, who are now reconciled. Both Helen and Menelaus also say that they returned to Sparta after a long voyage by way of Egypt. There, on the island of Pharos, Menelaus encounters the old sea-god Proteus, who tells him that Odysseus was a captive of the nymph Calypso. Telemachus learns the fate of Menelaus' brother, Agamemnon, king of Mycenae and leader of the Greeks at Troy: he was murdered on his return home by his wife Clytemnestra and her lover Aegisthus. The story briefly shifts to the suitors, who have only just realized that Telemachus is gone. Angry, they formulate a plan to ambush his ship and kill him as he sails back home. Penelope overhears their plot and worries for her son's safety. Escape to the Phaeacians (books 5–8) Charles Gleyre, Odysseus and Nausicaä  In the course of Odysseus' seven years as a captive of Calypso on the island Ogygia, she has fallen deeply in love with him, even though he spurns her offers of immortality as her husband and still mourns for home. She is ordered to release him by the messenger god Hermes, who has been sent by Zeus in response to Athena's plea. Odysseus builds a raft and is given clothing, food, and drink by Calypso. When Poseidon learns that Odysseus has escaped, he wrecks the raft, but helped by a veil given by the sea nymph Ino, Odysseus swims ashore on Scherie, the island of the Phaeacians. Naked and exhausted, he hides in a pile of leaves and falls asleep.  The next morning, awakened by girls' laughter, he sees the young Nausicaä, who has gone to the seashore with her maids after Athena told her in a dream to do so. He appeals for help. She encourages him to seek the hospitality of her parents, Arete and Alcinous. Alcinous promises to provide him a ship to return him home without knowing the identity of Odysseus. He remains for several days. Odysseus asks the blind singer Demodocus to tell the story of the Trojan Horse, a stratagem in which Odysseus had played a leading role. Unable to hide his emotion as he relives this episode, Odysseus at last reveals his identity. He then tells the story of his return from Troy.", "recordable": True}
        # This doesn't work
        _MAX_OPTION_SIZE = 1

        # Execution.
        test_big_val = sd._serialize_single_option(test_dict)


    @unittest.skipIf(not RUN_ALL_TESTS, "Temporarily disabled.")
    def test__serialize_single_option_small(self):
        ###
        # Configuration.
        test_dict = {"val": 3, "recordable": True}

        # Execution.
        test_small_val = sd._serialize_single_option(test_dict)

        # Validation.
        assert test_small_val == 3

    @unittest.skipIf(not RUN_ALL_TESTS, "Temporarily disabled.")
    def test__serialize_single_option_recordable(self):
        ###
        # Configuration.
        test_dict = {"val": 3}

        # Execution.
        test_recordable = sd._serialize_single_option(test_dict)

        # Validation.
        assert test_recordable == 'Not Recordable'

    @unittest.skipIf(not RUN_ALL_TESTS, "Temporarily disabled.")
    def test__serialize_single_option_undefined(self):
        ###
        # Configuration.
        test_dict = {"val": _UNDEFINED, "recordable": True}

        # Execution.
        test_undefined = sd._serialize_single_option(test_dict)

        # Validation.
        assert test_undefined == 'UNDEFINED'

    @unittest.skipIf(not RUN_ALL_TESTS, "Temporarily disabled.")
    def test_pull_connections_list(self):
        # (full_model_dict: dict) test_dict_basic
        ###
        # Configuration.
        test_dict = self.test_dict_basic

        # Execution.
        test_connections = sd.pull_connections_list(test_dict)
        print(f'test_connections: {test_connections}')

        # Validation.
        assert test_connections.sort() == ['R2', 'n1', 'n2', 'D1', 'R1'].sort()


    @unittest.skipIf( RUN_ALL_TESTS, "Temporarily disabled.")
    def test_model_to_dict(self):
        ###
        # Configuration.
        # Need better test that actually has partials

        # Execution.
        partial_list = sd._get_declare_partials(self.cir.model)
        json_vars = sd.model_to_dict(self.cir.model,
                          path=None,
                          values=_UNDEFINED,
                          case_id=None,
                          title="JDM")
        sd_name = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(sd_name, "reference_model.json"), "w") as f:
            json_reference = json.dump(f, json_vars)

        # Validation.
        json_reference = []
        # sd_name = "C:/Users/MUSIAKJ/Documents/ApMath/2024_work/NASA_MBSA/std_evaluator/standard_evaluator/src/system_description/"
        sd_name = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(sd_name, "reference_model.json"), "r") as f:
            json_reference = json.load(f)
        assert json_vars == json_reference



from standard_evaluator import OptProblem, EvaluatorInfo

def opt_problem_to_evaluator_info(opt_prob: OptProblem) -> EvaluatorInfo:
    return EvaluatorInfo(
        name=opt_prob.name,
        inputs=opt_prob.variables,
        outputs=opt_prob.responses,
        description=opt_prob.description,
        cite=opt_prob.cite,
        options=opt_prob.options,
    )


def evaluator_info_to_opt_problem(eval_info: EvaluatorInfo) -> OptProblem:
    return OptProblem(
        name=eval_info.name,
        variables=eval_info.inputs,
        responses=eval_info.outputs,
        description=eval_info.description,
        cite=eval_info.cite,
        options=eval_info.options,
    )


import copy
import pytest
import pandas as pd
import numpy as np
from standard_evaluator import FloatVariable, IntVariable, ArrayVariable
from standard_evaluator import EvaluatorInfo, OptProblem, MAXINT
from standard_evaluator import (
    evaluator_info_to_opt_problem,
    opt_problem_to_evaluator_info,
    CategoricalVariable,
)

def test_float_variable():
    my_float_var = FloatVariable(name="dummy")
    # Check that when no bounds are defined we set bounds to [-inf, inf]
    assert my_float_var.bounds[0] == -np.inf
    assert my_float_var.bounds[1] == np.inf
    with pytest.raises(ValueError):
        # Check that is lower bound is larger than upper bound we throw an error
        FloatVariable(name="dummy", bounds=[2.0, 1.0])
    # Check that if bounds are set they are saved correctly
    my_float_var2 = FloatVariable(name="dummy", bounds=[2.0, 4.0])
    assert my_float_var2.bounds[0] == 2.0
    assert my_float_var2.bounds[1] == 4.0

    with pytest.raises(ValueError):
        # Check that if scale is 0. we throw an error
        FloatVariable(name="dummy", scale=0.0)

    with pytest.raises(ValueError):
        # Check that if name contains spaces we throw an error
        FloatVariable(name="dummy 3")

    # Check that we can use an empty name
    my_float_var3 = FloatVariable(name="")
    assert my_float_var3.name == ""

def test_categorical_variable():
    my_cat_var = CategoricalVariable(name="x1", bounds=["x1 = 1", 4.3, 6.5])
    assert my_cat_var.bounds == ["x1 = 1", 4.3, 6.5]
    with pytest.raises(ValueError):
        # Check that is lower bound is larger than upper bound we throw an error
        IntVariable(name="dummy", bounds=[2.0, 1.0, "ret"], default="no")

def test_int_variable():
    my_int_var = IntVariable(name="dummy")
    # Check that when no bounds are defined we set bounds to [-inf, inf]
    assert my_int_var.bounds[0] == -MAXINT
    assert my_int_var.bounds[1] == MAXINT
    with pytest.raises(ValueError):
        # Check that is lower bound is larger than upper bound we throw an error
        IntVariable(name="dummy", bounds=[2.0, 1.0])
    # Check that if bounds are set they are saved correctly
    my_int_var2 = FloatVariable(name="dummy", bounds=[2, 4])
    assert my_int_var2.bounds[0] == 2
    assert my_int_var2.bounds[1] == 4

    with pytest.raises(ValueError):
        # Check that if scale is 0. we throw an error
        IntVariable(name="dummy", scale=0)


def test_array_variable():
    my_array_var = ArrayVariable(name="dummy", shape=(2, 3))
    # Check that when no bounds are defined we set bounds to [-inf, inf]
    assert np.all(np.isinf(my_array_var.bounds[0]))
    assert np.all(np.isinf(my_array_var.bounds[1]))
    # make sure the shift and scale are set correctly
    np.testing.assert_array_equal(my_array_var.shift, 0.0 * np.ones(shape=(2, 3)))
    np.testing.assert_array_equal(my_array_var.scale, 1.0 * np.ones(shape=(2, 3)))
    with pytest.raises(ValueError):
        # Check that is lower bound is larger than upper bound we throw an error
        ArrayVariable(name="dummy", bounds=[2.0, 1.0], shape=(3, 2))
    with pytest.raises(ValueError):
        # Check that either shape or default need to be defined
        ArrayVariable(name="dummy")
    # Check that if bounds are set they are saved correctly and expanded to the shape defined
    my_array_var2 = ArrayVariable(
        name="dummy", bounds=[2.0, 4.0], shape=(3, 2), shift=8.0, scale=0.1
    )
    np.testing.assert_array_equal(my_array_var2.bounds[0], 2.0 * np.ones(shape=(3, 2)))
    np.testing.assert_array_equal(my_array_var2.bounds[1], 4.0 * np.ones(shape=(3, 2)))
    # make sure the shift and scale are set correctly
    np.testing.assert_array_equal(my_array_var2.shift, 8.0 * np.ones(shape=(3, 2)))
    np.testing.assert_array_equal(my_array_var2.scale, 0.1 * np.ones(shape=(3, 2)))

    with pytest.raises(ValueError):
        # Check that if scale is 0. we throw an error
        ArrayVariable(name="dummy", scale=0.0, shape=(1, 2))

    # Check that if the default is given we take the shape from it
    default = np.array([[3.2, 6.3, 3.2], [1.2, 1.2, 34.4]])
    my_array_var3 = ArrayVariable(name="x4", default=default, scale=2.3)
    assert my_array_var3.shape == (2, 3)

    with pytest.raises(ValueError):
        # We cannot set the scale to a different shape than default
        ArrayVariable(name="x4", default=default, scale=np.array([2.3]))

    with pytest.raises(ValueError):
        # We cannot set the shift to a different shape than default
        ArrayVariable(name="x4", default=default, shift=np.array([2.3]))

    with pytest.raises(ValueError):
        # We cannot set the bounds to a different shape than default
        ArrayVariable(
            name="x4", default=default, bounds=[np.array([2.3]), np.array([2.3])]
        )

    with pytest.raises(ValueError):
        # We cannot set the shape to a different shape than default
        ArrayVariable(name="x4", default=default, shape=(2,))

    my_array_var4 = ArrayVariable(
        name="dummy",
        default=default,
        bounds=[-default, default],
        shift=8.0 * default,
        scale=0.1 * default,
    )
    np.testing.assert_array_equal(my_array_var4.bounds[0], -default)
    np.testing.assert_array_equal(my_array_var4.bounds[1], default)
    # make sure the shift and scale are set correctly
    np.testing.assert_array_equal(my_array_var4.shift, 8.0 * default)
    np.testing.assert_array_equal(my_array_var4.scale, 0.1 * default)

    with pytest.raises(ValueError):
        # Check that if any element of scale is 0. we throw an error
        ArrayVariable(name="dummy", scale=np.array([3.2, 0.0]), shape=(1, 2))


def test_opt_problem():
    with pytest.raises(ValueError):
        # Check that if if you try to name two inputs the same it throws an error
        OptProblem(name="bla", variables=[{"name": "x0"}, {"name": "x0"}])

    with pytest.raises(ValueError):
        # Check that if if you try to name two outputs the same it throws an error
        OptProblem(
            name="bla",
            variables=[{"name": "x0"}],
            responses=[{"name": "x0"}, {"name": "x0"}],
        )

    with pytest.raises(ValueError):
        # We need to define at least one input
        OptProblem(name="bla", variables=[])

    variables = [
        ArrayVariable(name="x1", default=[[299.3, 243.5]]),
        FloatVariable(name="p2", class_type="float"),
    ]
    responses = [FloatVariable(name="y2", class_type="float")]
    my_prob = OptProblem(
        variables=variables, responses=responses, objectives=["y2"], constraints=["y2"]
    )
    expected = {
        "name": "opt_problem",
        "class_type": "OptProblem",
        "variables": [
            {
                "name": "x1",
                "default": np.array([[299.3, 243.5]]),
                "bounds": None,
                "shift": None,
                "scale": None,
                "units": None,
                "description": "",
                "class_type": "floatarray",
                "shape": (1, 2),
                "options": {},
            },
            {
                "name": "p2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            },
        ],
        "responses": [
            {
                "name": "y2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            }
        ],
        "objectives": ["y2"],
        "constraints": ["y2"],
        "description": None,
        "cite": None,
        "options": {},
    }
    np.testing.assert_equal(expected, my_prob.model_dump())

    my_prob = OptProblem(
        name="dummy",
        variables=variables,
        responses=responses,
        objectives=["y2"],
        constraints=["y2"],
    )
    expected = {
        "name": "dummy",
        "class_type": "OptProblem",
        "variables": [
            {
                "name": "x1",
                "default": np.array([[299.3, 243.5]]),
                "bounds": None,
                "shift": None,
                "scale": None,
                "units": None,
                "description": "",
                "class_type": "floatarray",
                "shape": (1, 2),
                "options": {},
            },
            {
                "name": "p2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            },
        ],
        "responses": [
            {
                "name": "y2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            }
        ],
        "objectives": ["y2"],
        "constraints": ["y2"],
        "description": None,
        "cite": None,
        "options": {},
    }
    np.testing.assert_equal(expected, my_prob.model_dump())

    with pytest.raises(ValueError):
        # A variable cannot be a constraint
        OptProblem(
            variables=variables,
            responses=responses,
            objectives=["y2"],
            constraints=["p2"],
        )

    with pytest.raises(ValueError):
        # An element in an array variable cannot be outside the shape
        OptProblem(
            variables=variables,
            responses=responses,
            objectives=["x1[0,3]"],
            constraints=["y2"],
        )

    with pytest.raises(ValueError):
        # You cannot have a constraint that is not a response
        OptProblem(
            variables=variables,
            responses=responses,
            objectives=["y2"],
            constraints=["Hello39"],
        )

    my_prob3 = OptProblem(
        name="dummy",
        variables=variables,
        responses=responses,
        objectives=["x1[0,1]"],
        constraints=["y2"],
    )
    expected = {
        "name": "dummy",
        "class_type": "OptProblem",
        "variables": [
            {
                "name": "x1",
                "default": np.array([[299.3, 243.5]]),
                "bounds": None,
                "shift": None,
                "scale": None,
                "units": None,
                "description": "",
                "class_type": "floatarray",
                "shape": (1, 2),
                "options": {},
            },
            {
                "name": "p2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            },
        ],
        "responses": [
            {
                "name": "y2",
                "default": None,
                "bounds": (-np.inf, np.inf),
                "shift": 0.0,
                "scale": 1.0,
                "units": None,
                "description": "",
                "class_type": "float",
                "options": {},
            }
        ],
        "objectives": ["x1[0,1]"],
        "constraints": ["y2"],
        "description": None,
        "cite": None,
        "options": {},
    }
    np.testing.assert_equal(expected, my_prob3.model_dump())
